-- utility function module for scenario scripting
-- Usage:
-- Use lib. prefix for utility and support functions:
-- lib = require "shared"	-- file shared.lua in parent campaign directory

Mshared = {}	-- Module table (needs to be a global, for re-loads ... defect in UE4 Lua plugin)

-- naming convention:	functions starting with Module prefix, initial upper-case, and CamelCased names are "public" and exported
-- 						functions and variables not prefixed, initial lower-case, and snake_cased names are "private" and module-internal

-- UE4 Lua plugin work-around: re-loads do not re-load "require"d modules, so each module function has a re-load guard for that case
-- Also, the Mshared variable can _not_ be local, because the re-load failure causes a nil _ENV (preventing use of global functions)

----------------------------------------------------
-- Utility functions: generic Lua table searching --
----------------------------------------------------

function Mshared.FindItemInTable(atable, item)	-- returns integer index of item in table, or -1 if not found
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	for index,value in pairs(atable) do
		if (value == item) then return index end	-- found
	end
	return -1	-- not found
end

function Mshared.IsItemInTable(atable, item)	-- returns bool indicating whether item was found in table
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	return Mshared.FindItemInTable(atable,item) ~= -1
end

-----------------------------------
-- Proximity/Encounter detection --
-----------------------------------

function Mshared.Distance(pos1, pos2)	-- returns the Cartesian distance (as number of Hexes) between the given {x,y} positions as a positive number
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local dx = pos2[1] - pos1[1]
	local dy = pos2[2] - pos1[2]
	local p = 0
	if ((dx * dy) < 0) then p = 1 end
	local free = math.floor((pos1[2]+p)/2) - math.floor((pos2[2]+p)/2)
	return math.abs(dy) + math.max(math.abs(dx)-math.abs(free),0)
end

-------------------------------------------
-- Support functions, using PzC2 Lua API --
-------------------------------------------
-- Assumes PzC2 Lua API is available.  See: PanzerCorps2/Content/Editor/properties.txt

--[[
	The filter argument is a table with values for one or more of the following keys:
	.player		id to match vs unit.owner_id
	.ground		boolean, to match vs unit.type:IsLandUnit()
	.air		boolean, to match vs unit.type:IsAircraft()
	.naval		boolean, to match vs unit.type:IsNaval()
	.class		table of UnitClas. value(s) to match vs unit.type.kind
	.id			table of unit ids to match vs unit.id
	.spotted_by	player_id to check vs unit.spotted

	Values to use for .class filter:
	ground_classes=	{UnitClass.Infantry, UnitClass.Tank, UnitClass.Recon, UnitClass.AntiTank, UnitClass.Artillery, UnitClass.AntiAir,
					 UnitClass.Structure, UnitClass.ArmoredTrain, UnitClass.LandTransport, UnitClass.RailTransport}
	air_classes=	{UnitClass.Fighter, UnitClass.TacticalBomber, UnitClass.StrategicBomber, UnitClass.AirTransport}
	naval_classes=	{UnitClass.Submarine, UnitClass.Destroyer, UnitClass.CapitalShip, UnitClass.Carrier, UnitClass.SeaTransport}
]]
function Mshared.FindUnits(filter)	-- returns table of Units matching criteria
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local result = {}
	for _,unit in pairs(world.units) do
		if (filter.player ~= nil and unit.owner_id ~= filter.player) then goto continue end
		if (filter.ground ~= nil and unit.type:IsLandUnit() ~= filter.ground) then goto continue end
		if (filter.air ~= nil and unit.type:IsAircraft() ~= filter.air) then goto continue end
		if (filter.naval ~= nil and unit.type:IsNaval() ~= filter.naval) then goto continue end
		if (filter.class ~= nil and not Mshared.IsItemInTable(filter.class, unit.type.kind)) then goto continue end
		if (filter.id ~= nil and not Mshared.IsItemInTable(filter.id, unit.id)) then goto continue end
		if (filter.spotted_by ~= nil and (unit.spotted & (1 << filter.spotted_by)) == 0) then goto continue end
		table.insert(result, unit)
		::continue::
	end
	return result
end

function Mshared.ProximityCheck(filter1, filter2, distance)	-- returns boolean indicating whether any of the Units selected by the given filters (see: FindUnits) are withing the given distance (in Hexes, see: Distance)
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local min_dist = distance * 2	-- arbitrarily large
	for _,unit1 in pairs(Mshared.FindUnits(filter1)) do	-- for all Units meeting 1st filter
		for _,unit2 in pairs(Mshared.FindUnits(filter2)) do	-- and all Units meeting 2nd filter
			min_dist = math.min(min_dist,Mshared.Distance(unit1.position,unit2.position))
			if (min_dist <= distance) then return true end	-- early loop escape, if both filter sets are large
		end
	end
	return false	-- reached end of loops without meeting threshold
end

function Mshared.InitCommendationPoints(value)	-- UI: returns the given number value, also using it to set up player 0 HUD for CP display (based on player.resource1* values)
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local player = world:GetPlayer(0)	-- PzC2 Lua API
	player.resource1 = value or 0	-- player.resource1/2/3 is a number, resource 1 used for CP
	player.resource1_icon = "/Game/Gui/GameplayUI/hud_commendation_icon"
	player.resource1_tooltip= NSLOCTEXT("common", "commendation_points", "Commendation Points")
	return player.resource1
end

function Mshared.UpdateCommendationPoints(comm)	-- returns the CP value updated by the given number value, also updating player.resource1 value
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local player = world:GetPlayer(0)	-- PzC2 Lua API
	player.resource1 = math.max(0,player.resource1 + comm)	-- {ensure >= 0} player.resource1 tracks the updated CP value
	return player.resource1
end

function Mshared.AllySetOrder(player_id, order)	-- sets Ally AI commands for the given player_id, order one of: "AIOrderAttacker", "AIOrderDefender", "AIOrderHoldPositionFire"
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local ally = player_id or 1	-- default "Ally" player id
	for _,u in pairs(world.units) do	-- for all units of that player:
		if (u.owner_id == ally) then
			if (next(u.orders) == nil or u.orders[1]:GetOrderClass() == "AIOrderDefault") then	-- replace any existing orders:
				u.orders = {NewOrder(order)}	-- PzC2 Lua API
			end
		end
	end
end

-- add a single item of equipment
-- Possible category values: Prototypes, Captured, LimitedAvailability
-- NOTE: do not put other nation equipment into Prototypes
function Mshared.AddEquipment(player_id, unit_type, quantity, category)	-- adds the given quantity of the given UnitType to player.equipment table
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local player = world:GetPlayer(player_id)	-- PzC2 Lua API
	local eqtype = GetUnitType(unit_type)	-- PzC2 Lua API
	-- PzC2 Lua API _really_ dislikes update-in-place.  Use bulk replace:
	local equipment = player.equipment
	local eq = equipment[eqtype]
	if (eq == nil) then	-- 1st time storing this eqtype
		eq = NewEquipment()	-- PzC2 Lua API
		eq.category = category	-- @TODO: error-check value
		equipment[eqtype] = eq
	end
	-- assumes eq.category is same as previously set when initialized
	eq.quantity = eq.quantity + quantity
	player.equipment = equipment
end

-- add a table of equipment
-- Possible category values: Prototypes, Captured, LimitedAvailability
-- NOTE: do not put other nation equipment into Prototypes
function Mshared.AddEquipments(player_id, equips, category)	-- adds the given table (of {unit_type,quantity} pairs) to the player.equipment in the given category
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	if (#equips > 0) then
		for _,eq in pairs(equips) do
			Mshared.AddEquipment(player_id, eq[1], eq[2], category)
		end
	end
end

function Mshared.CountEquipment(player_id, unit_type)	-- returns the number of a given UnitType in the given player's .equipment table
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local player = world:GetPlayer(player_id)
	local equipment = player.equipment
	local eqtype = GetUnitType(unit_type)
	local eq = equipment[eqtype]
	if (eq == nil) then	-- not found
		return 0
	else
		return eq.quantity
	end
end

function Mshared.CountSide1VHs(VHs)	-- return integer count of hexes in VHs (table of Point|Hex) owned by side 0 (human player's side)
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local n = 0
	for _,vh in ipairs(VHs) do
		local hex = world:GetHex(vh)	-- PzC2 Lua API
		local owner = world:GetPlayer(hex.owner)
		local side = owner.side_id
		if (side == 0) then n = n + 1 end	-- count it
	end
	return n
end

function Mshared.CountUnitsEntered(unittype, zone) -- returns count of Units of given type in given Hex zone
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local n = 0
	for _, hex in pairs(zone) do
		for _,unit in pairs(world:GetHex(hex):GetUnits()) do	-- PzC2 Lua API
			if (unittype == nil or unit.type.kind == unittype) then n = n + 1 end	-- count it
		end
	end
	return n
end

-- The given u table is in the following format: {"<Unit ID>", "<Transport ID>", overstrength, experience}
-- If no organic transport is needed, its respective field should be empty string ("").
function Mshared.CreateUnit(player_id,u,aux)	-- returns Unit created with given player_id, u table, and optional auxiliary flag, or nil on error
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local owner = world:GetPlayer(player_id)	-- PzC2 Lua API
	-- 1. Create and execute purchase action, use unit type, transport type and overstrength as parameters
	local create_action = world:MakePurchaseAction(owner, u[1], u[2], u[3])	-- PzC2 Lua API
	if (create_action == nil) then return nil end
	if (aux ~= nil and aux == true) then create_action.auxiliary = true else create_action.auxiliary = false end
	create_action.cost = 0 -- don't take money for units we generate
	create_action.faction = owner.factions[1]  --set to first faction of owning player for created units
	world:Exec(create_action)	-- PzC2 Lua API

	-- 2. Create and execute deploy action, place the unit in the vacant hex
	local unit = world:GetUnit(create_action.id)	-- PzC2 Lua API
	if (unit == nil) then return nil end
	-- 3. Modify the unit to set its experience to the value supplied
	unit.experience = u[4]
	-- 4. Modify the unit to set its aggressiveness to the value supplied
	player_aggressiveness = {0, 85, 70, 70}	-- @FIXME: read these from Player::AISettings defined in scenario
	unit.aggressiveness = player_aggressiveness[player_id+1]
	return unit
end

function Mshared.FindEmptyHex(zone)	-- returns empty Hex in given zone (table of {x,y} Point/Hex pairs), or nil if none available
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	for _,h in pairs(zone) do
		local hex = world:GetHex(h)	-- PzC2 Lua API
		if (hex:GetUnit(0) == nil) then return hex end -- found empty Hex
	end
	return nil	-- not found
end

function Mshared.DeployUnit(unit,hex)	-- deploys the given Unit in the given Hex
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	-- MakeDeployAction expects Point, not Hex, as 2nd argument:
	if (unit ~= nil and hex ~= nil) then world:Exec(world:MakeDeployAction(unit,hex.index)) end	-- PzC2 Lua API
end

-- The units argument is a list of tables in the following format: {"<Unit ID>", "<Transport ID>", overstrength, experience}
-- If no organic transport is needed, its respective field should be empty string ("").
function Mshared.SpawnWave(player_id,zone,units,aux)	-- returns table of Unit.id after creating given Units (optionally, as auxiliary) for given player_id and deploys them in vacant Hexes in given zone
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	local spawned_ids = {}
	for _,u in ipairs(units) do
		-- Find the next vacant hex in the zone
		local hex = Mshared.FindEmptyHex(zone)
		if (hex == nil) then return end		-- have run out of empty Hexes in zone, no need to continue looking
		-- Create the unit in the free hex we've found
		local unit = Mshared.CreateUnit(player_id,u,aux)
		Mshared.DeployUnit(unit,hex)
		table.insert(spawned_ids,unit.id)
	end
	return spawned_ids
end

----------------------------
-- Hero-related functions --
----------------------------

local function is_empty_string(str)	-- returns boolean indicating whether given string (or FText) is nil or empty
	local s = tostring(str)
	return str == nil or #s == 0 or s == "" or s == "nil"	-- 3rd & 4th tests probably redundant
end

local function name_matches(name,s)	-- returns boolean indicating whether given name (non-empty string or FText) matches given non-empty string (or FText)
	local s1 = tostring(name)
	local s2 = tostring(s)
	return not is_empty_string(s1) and not is_empty_string(s2) and #s1 == #s2 and string.find(s1,s2,1,true) ~= nil
end

-- multi-value return: hero,unit (either/both may be nil)
function Mshared.FindHeroByName(player_id, name)	-- returns given player_id's Hero with given name, or nil if not found, and Unit if assigned or nil if not
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	if (not is_empty_string(name)) then	-- only check non-empty names
		-- check un-assigned Heroes:
		for _,hero in pairs(world:GetPlayer(player_id).heroes) do 	-- assume valid player_id and Player object returned
			if (name_matches(hero.name,name)) then return hero,nil end	-- Hero found, no Unit
		end
		--	check all Units for possible Hero assignment:
		for _,unit in pairs(world.units) do	-- for all Units
			if (unit.owner_id == player_id) then	-- that belong to player
				for _,hero in pairs(unit:GetHeroes()) do	-- assume nil table is harmless	-- PzC2 Lua API
					if (name_matches(hero.name,name)) then return hero,unit end	-- found Hero and assigned Unit
				end
			end
		end
	end
	return nil,nil	-- not found
end

-- expects: valid owner_id, heroes_table attributes of: unit_id, optional/defaulted name (NSLOCTEXT FText), traits (table of UnitTrait. value(s)), *modifiers, unit_classes, and portrait (string: path in /Game/Gui/Common/Heroes)
function Mshared.CreateAndAssignHeroes(player_id, heroes_table, silent)	-- generates and assigns Heroes for the given player_id with the given table of attributes, optionally silent about it
	if (not package.loaded["shared"]) then Mshared = require "shared" end	-- re-load guard
	for _,h in pairs(heroes_table) do
		local hero, unit	-- both start nil to indicate not found
		local existing_hero = false
		local hero_already_assigned = false

		-- if non-empty h.name supplied, check for existing Hero:
		if (not is_empty_string(h.name)) then
			hero,unit = Mshared.FindHeroByName(player_id,h.name)	-- one or both may come back nil
			existing_hero = (hero ~= nil)
			hero_already_assigned = (unit ~= nil)
		end
		if (not existing_hero) then	-- create a new Hero:
			hero = NewHero()	-- PzC2 Lua API
			if (not is_empty_string(h.name)) then hero.name = h.name end	-- copy FText
			hero.portrait = h.portrait or "/Game/Gui/Common/Heroes/generic_standin.generic_standin"
			-- traits:
			hero.extra_traits = h.traits or {UnitTrait.TenaciousDefender}	-- always have something
			if (h.extra_traits ~= nil) then hero.extra_traits = h.extra_traits end	-- cope with typo: extra_traits= (rather than traits=)
			-- Assume NewHero() assigns no modifiers:
			if (h.modifiers ~= nil) then hero.modifiers = h.modifiers end
			if (h.attack_modifiers ~= nil) then hero.attack_modifiers = h.attack_modifiers end
			if (h.defense_modifiers ~= nil) then hero.defense_modifiers = h.defense_modifiers end
			if (h.unit_classes ~= nil) then hero.unit_classes = h.unit_classes end

			-- Fix for caller supplying incorrect player_id, but needing to assign to Unit (UE4 hang/crash):
			if (unit ~= nil and player_id ~= unit.owner_id) then player_id = unit.owner_id end
			local action = world:MakeNewHeroAction(player_id, hero)	-- PzC2 Lua API
			if (silent ~= nil and silent == true) then action.silent = true end	-- Wagner, mine heroes
			world:Exec(action)	-- PzC2 Lua API
		end
		-- valid Hero (and hero.id) at this point, continue on with assigning to Unit (or not):

		-- if h.unit_id given, then (possibly re-)assign Hero iff Unit exists, else just leave Hero as is:
		if (h.unit_id == nil) then return end
		if (existing_hero and hero_already_assigned and unit.id ~= h.unit_id) then	-- Hero is currently assigned to another Unit, unassign:
			-- (in all known use cases, this isn't required, but is a good defensive check)
			world:Exec(world:MakeUnassignHeroAction(unit, hero.id))	-- PzC2 Lua API
			unit = nil	-- no longer a valid association
		end
		-- at this point, Hero is either unassigned or already assigned to the right Unit:
		hero_already_assigned = (unit ~= nil)	-- evaluates "false" in all cases seen so far
		if (unit == nil) then unit = world:GetUnit(h.unit_id) end	-- PzC2 Lua API
		if (unit ~= nil) then	-- assign Hero to Unit:
			-- if h.name not supplied (or empty ""), pick up hero.name from unit.name (if any) else just give generic name:
			local str = string.format("Generic Hero #%d",hero.id+1)	-- @TODO: NSLOCTEXT("common", "EnemyHero", "Generic Hero")
			if (is_empty_string(hero.name)) then hero.name = str end
			-- 2nd test is a hack, until there's better UE4 Lua API support for UE4 FText string (i.e. Unicode support):
			if (not is_empty_string(unit.name) and tostring(unit.name) ~= "%num %type") then hero.name = unit.name end	-- copy FText
			if (not hero_already_assigned) then world:Exec(world:MakeAssignHeroAction(unit, hero.id)) end	-- PzC2 Lua API
		end
	end
end


return Mshared
